@@ -1,61 +1,26 @@
+import gzip
 import json as json_module
+import re
 from json.decoder import JSONDecodeError
 from typing import Any
 from typing import Callable
-from typing import Dict
 from typing import List
+from typing import Mapping
+from typing import MutableMapping
 from typing import Optional
+from typing import Pattern
 from typing import Tuple
 from typing import Union
 from urllib.parse import parse_qsl
 from urllib.parse import urlparse
 
 from requests import PreparedRequest
-from requests.packages.urllib3.util.url import parse_url
-
-
-def _create_key_val_str(input_dict: Union[Dict[Any, Any], Any]) -> str:
-    """
-    Returns string of format {'key': val, 'key2': val2}
-    Function is called recursively for nested dictionaries
-
-    :param input_dict: dictionary to transform
-    :return: (str) reformatted string
-    """
-
-    def list_to_str(input_list: List[str]) -> str:
-        """
-        Convert all list items to string.
-        Function is called recursively for nested lists
-        """
-        converted_list = []
-        for item in sorted(input_list, key=lambda x: str(x)):
-            if isinstance(item, dict):
-                item = _create_key_val_str(item)
-            elif isinstance(item, list):
-                item = list_to_str(item)
-
-            converted_list.append(str(item))
-        list_str = ", ".join(converted_list)
-        return "[" + list_str + "]"
-
-    items_list = []
-    for key in sorted(input_dict.keys(), key=lambda x: str(x)):
-        val = input_dict[key]
-        if isinstance(val, dict):
-            val = _create_key_val_str(val)
-        elif isinstance(val, list):
-            val = list_to_str(input_list=val)
-
-        items_list.append("{}: {}".format(key, val))
-
-    key_val_str = "{{{}}}".format(", ".join(items_list))
-    return key_val_str
+from urllib3.util.url import parse_url
 
 
 def _filter_dict_recursively(
-    dict1: Dict[Any, Any], dict2: Dict[Any, Any]
-) -> Dict[Any, Any]:
+    dict1: Mapping[Any, Any], dict2: Mapping[Any, Any]
+) -> Mapping[Any, Any]:
     filtered_dict = {}
     for k, val in dict1.items():
         if k in dict2:
@@ -67,7 +32,7 @@ def _filter_dict_recursively(
 
 
 def urlencoded_params_matcher(
-    params: Optional[Dict[str, str]], *, allow_blank: bool = False
+    params: Optional[Mapping[str, str]], *, allow_blank: bool = False
 ) -> Callable[..., Any]:
     """
     Matches URL encoded data
@@ -87,8 +52,8 @@ def urlencoded_params_matcher(
         params_dict = params or {}
         valid = params is None if request_body is None else params_dict == qsl_body
         if not valid:
-            reason = "request.body doesn't match: {} doesn't match {}".format(
-                _create_key_val_str(qsl_body), _create_key_val_str(params_dict)
+            reason = (
+                f"request.body doesn't match: {qsl_body} doesn't match {params_dict}"
             )
 
         return valid, reason
@@ -97,7 +62,7 @@ def urlencoded_params_matcher(
 
 
 def json_params_matcher(
-    params: Optional[Union[Dict[str, Any], List[Any]]], *, strict_match: bool = True
+    params: Optional[Union[Mapping[str, Any], List[Any]]], *, strict_match: bool = True
 ) -> Callable[..., Any]:
     """Matches JSON encoded data of request body.
 
@@ -124,8 +89,11 @@ def json_params_matcher(
         request_body = request.body
         json_params = (params or {}) if not isinstance(params, list) else params
         try:
-            if isinstance(request_body, bytes):
-                request_body = request_body.decode("utf-8")
+            if isinstance(request.body, bytes):
+                try:
+                    request_body = request.body.decode("utf-8")
+                except UnicodeDecodeError:
+                    request_body = gzip.decompress(request.body).decode("utf-8")
             json_body = json_module.loads(request_body) if request_body else {}
 
             if (
@@ -139,13 +107,7 @@ def json_params_matcher(
             valid = params is None if request_body is None else json_params == json_body
 
             if not valid:
-                if isinstance(json_body, dict) and isinstance(json_params, dict):
-                    reason = "request.body doesn't match: {} doesn't match {}".format(
-                        _create_key_val_str(json_body), _create_key_val_str(json_params)
-                    )
-                else:
-                    reason = f"request.body doesn't match: {json_body} doesn't match {json_params}"
-
+                reason = f"request.body doesn't match: {json_body} doesn't match {json_params}"
                 if not strict_match:
                     reason += (
                         "\nNote: You use non-strict parameters check, "
@@ -186,7 +148,7 @@ def fragment_identifier_matcher(identifier: Optional[str]) -> Callable[..., Any]
 
 
 def query_param_matcher(
-    params: Optional[Dict[str, Any]], *, strict_match: bool = True
+    params: Optional[MutableMapping[str, Any]], *, strict_match: bool = True
 ) -> Callable[..., Any]:
     """Matcher to match 'params' argument in request.
 
@@ -227,10 +189,7 @@ def query_param_matcher(
         valid = sorted(params_dict.items()) == sorted(request_params_dict.items())
 
         if not valid:
-            reason = "Parameters do not match. {} doesn't match {}".format(
-                _create_key_val_str(request_params_dict),
-                _create_key_val_str(params_dict),
-            )
+            reason = f"Parameters do not match. {request_params_dict} doesn't match {params_dict}"
             if not strict_match:
                 reason += (
                     "\nYou can use `strict_match=True` to do a strict parameters check."
@@ -251,7 +210,7 @@ def query_string_matcher(query: Optional[str]) -> Callable[..., Any]:
 
     def match(request: PreparedRequest) -> Tuple[bool, str]:
         reason = ""
-        data = parse_url(request.url)
+        data = parse_url(request.url or "")
         request_query = data.query
 
         request_qsl = sorted(parse_qsl(request_query)) if request_query else {}
@@ -260,9 +219,9 @@ def query_string_matcher(query: Optional[str]) -> Callable[..., Any]:
         valid = not query if request_query is None else request_qsl == matcher_qsl
 
         if not valid:
-            reason = "Query string doesn't match. {} doesn't match {}".format(
-                _create_key_val_str(dict(request_qsl)),
-                _create_key_val_str(dict(matcher_qsl)),
+            reason = (
+                "Query string doesn't match. "
+                f"{dict(request_qsl)} doesn't match {dict(matcher_qsl)}"
             )
 
         return valid, reason
@@ -270,7 +229,7 @@ def query_string_matcher(query: Optional[str]) -> Callable[..., Any]:
     return match
 
 
-def request_kwargs_matcher(kwargs: Optional[Dict[str, Any]]) -> Callable[..., Any]:
+def request_kwargs_matcher(kwargs: Optional[Mapping[str, Any]]) -> Callable[..., Any]:
     """
     Matcher to match keyword arguments provided to request
 
@@ -292,8 +251,8 @@ def request_kwargs_matcher(kwargs: Optional[Dict[str, Any]]) -> Callable[..., An
         )
 
         if not valid:
-            reason = "Arguments don't match: {} doesn't match {}".format(
-                _create_key_val_str(request_kwargs), _create_key_val_str(kwargs_dict)
+            reason = (
+                f"Arguments don't match: {request_kwargs} doesn't match {kwargs_dict}"
             )
 
         return valid, reason
@@ -302,7 +261,7 @@ def request_kwargs_matcher(kwargs: Optional[Dict[str, Any]]) -> Callable[..., An
 
 
 def multipart_matcher(
-    files: Dict[str, Any], data: Optional[Dict[str, str]] = None
+    files: Mapping[str, Any], data: Optional[Mapping[str, str]] = None
 ) -> Callable[..., Any]:
     """
     Matcher to match 'multipart/form-data' content-type.
@@ -386,7 +345,7 @@ def multipart_matcher(
 
 
 def header_matcher(
-    headers: Dict[str, str], strict_match: bool = False
+    headers: Mapping[str, Union[str, Pattern[str]]], strict_match: bool = False
 ) -> Callable[..., Any]:
     """
     Matcher to match 'headers' argument in request using the responses library.
@@ -402,18 +361,36 @@ def header_matcher(
     :return: (func) matcher
     """
 
+    def _compare_with_regex(request_headers: Union[Mapping[Any, Any], Any]) -> bool:
+        if strict_match and len(request_headers) != len(headers):
+            return False
+
+        for k, v in headers.items():
+            if request_headers.get(k) is not None:
+                if isinstance(v, re.Pattern):
+                    if re.match(v, request_headers[k]) is None:
+                        return False
+                else:
+                    if not v == request_headers[k]:
+                        return False
+            else:
+                return False
+
+        return True
+
     def match(request: PreparedRequest) -> Tuple[bool, str]:
-        request_headers: Union[Dict[Any, Any], Any] = request.headers or {}
+        request_headers: Union[Mapping[Any, Any], Any] = request.headers or {}
 
         if not strict_match:
             # filter down to just the headers specified in the matcher
             request_headers = {k: v for k, v in request_headers.items() if k in headers}
 
-        valid = sorted(headers.items()) == sorted(request_headers.items())
+        valid = _compare_with_regex(request_headers)
 
         if not valid:
-            return False, "Headers do not match: {} doesn't match {}".format(
-                _create_key_val_str(request_headers), _create_key_val_str(headers)
+            return (
+                False,
+                f"Headers do not match: {request_headers} doesn't match {headers}",
             )
 
         return valid, ""
