@@ -1,3 +1,7 @@
+import gzip
+import re
+from typing import Any
+from typing import List
 from unittest.mock import Mock
 
 import pytest
@@ -131,13 +135,27 @@ def test_json_params_matcher_not_strict_diff_values():
             )
         assert (
             "- POST http://example.com/ request.body doesn't match: "
-            "{page: {type: json}} doesn't match {page: {diff: value, type: json}}"
+            "{'page': {'type': 'json'}} doesn't match {'page': {'type': 'json', 'diff': 'value'}}"
         ) in str(exc.value)
 
     run()
     assert_reset()
 
 
+def test_failed_matchers_dont_modify_inputs_order_in_error_message():
+    json_a = {"array": ["C", "B", "A"]}
+    json_b = '{"array" : ["B", "A", "C"]}'
+    mock_request = Mock(body=json_b)
+    result = matchers.json_params_matcher(json_a)(mock_request)
+    assert result == (
+        False,
+        (
+            "request.body doesn't match: {'array': ['B', 'A', 'C']} "
+            "doesn't match {'array': ['C', 'B', 'A']}"
+        ),
+    )
+
+
 def test_json_params_matcher_json_list():
     json_a = [{"a": "b"}]
     json_b = '[{"a": "b", "c": "d"}]'
@@ -150,13 +168,21 @@ def test_json_params_matcher_json_list():
 
 
 def test_json_params_matcher_json_list_empty():
-    json_a = []
+    json_a: "List[Any]" = []
     json_b = "[]"
     mock_request = Mock(body=json_b)
     result = matchers.json_params_matcher(json_a)(mock_request)
     assert result == (True, "")
 
 
+def test_json_params_matcher_body_is_gzipped():
+    json_a = {"foo": 42, "bar": None}
+    json_b = gzip.compress(b'{"foo": 42, "bar": null}')
+    mock_request = Mock(body=json_b)
+    result = matchers.json_params_matcher(json_a)(mock_request)
+    assert result == (True, "")
+
+
 def test_urlencoded_params_matcher_blank():
     @responses.activate
     def run():
@@ -238,7 +264,7 @@ def test_query_param_matcher_loose_fail():
 
         assert (
             "- GET https://example.com/ Parameters do not match. {} doesn't"
-            " match {does_not_exist: test}\n"
+            " match {'does_not_exist': 'test'}\n"
             "You can use `strict_match=True` to do a strict parameters check."
         ) in str(exc.value)
 
@@ -292,7 +318,7 @@ def test_request_matches_empty_body():
                 )
 
             msg = str(excinfo.value)
-            assert "request.body doesn't match: {my: data} doesn't match {}" in msg
+            assert "request.body doesn't match: {'my': 'data'} doesn't match {}" in msg
 
         with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
             rsps.add(
@@ -310,9 +336,9 @@ def test_request_matches_empty_body():
                 )
             msg = str(excinfo.value)
             assert (
-                "request.body doesn't match: {page: second, type: urlencoded} doesn't match {}"
-                in msg
-            )
+                "request.body doesn't match: {'page': 'second', "
+                "'type': 'urlencoded'} doesn't match {}"
+            ) in msg
 
     run()
     assert_reset()
@@ -377,7 +403,8 @@ def test_fail_matchers_error():
 
             msg = str(excinfo.value)
             assert (
-                "request.body doesn't match: {id: bad} doesn't match {foo: bar}" in msg
+                "request.body doesn't match: {'id': 'bad'} doesn't match {'foo': 'bar'}"
+                in msg
             )
 
             assert (
@@ -406,10 +433,11 @@ def test_fail_matchers_error():
 
             msg = str(excinfo.value)
             assert (
-                "Parameters do not match. {id: bad} doesn't match {my: params}" in msg
+                "Parameters do not match. {'id': 'bad'} doesn't match {'my': 'params'}"
+                in msg
             )
             assert (
-                "request.body doesn't match: {page: two} doesn't match {page: one}"
+                "request.body doesn't match: {'page': 'two'} doesn't match {'page': 'one'}"
                 in msg
             )
 
@@ -430,7 +458,7 @@ def test_fail_matchers_error():
             msg = str(excinfo.value)
             assert (
                 "Arguments don't match: "
-                "{stream: True, verify: True} doesn't match {stream: True, verify: False}"
+                "{'stream': True, 'verify': True} doesn't match {'stream': True, 'verify': False}"
             ) in msg
 
     run()
@@ -447,7 +475,7 @@ def test_fail_matchers_error():
         (b"\xacHello World!", b"\xacHello World!"),
     ],
 )
-def test_multipart_matcher(req_file, match_file):
+def test_multipart_matcher(req_file, match_file):  # type: ignore[misc]
     @responses.activate
     def run():
         req_data = {"some": "other", "data": "fields"}
@@ -575,9 +603,9 @@ def test_query_string_matcher_raises():
 
             msg = str(excinfo.value)
             assert (
-                "Query string doesn't match. {didi: pro, test: 1} doesn't match {didi: pro}"
-                in msg
-            )
+                "Query string doesn't match. {'didi': 'pro', 'test': '1'} "
+                "doesn't match {'didi': 'pro'}"
+            ) in msg
 
     run()
     assert_reset()
@@ -614,7 +642,7 @@ def test_request_matches_headers():
     assert_reset()
 
 
-def test_request_matches_headers_no_match():
+def test_request_header_value_mismatch_raises():
     @responses.activate
     def run():
         url = "http://example.com/"
@@ -630,8 +658,31 @@ def test_request_matches_headers_no_match():
 
         msg = str(excinfo.value)
         assert (
-            "Headers do not match: {Accept: application/xml} doesn't match "
-            "{Accept: application/json}"
+            "Headers do not match: {'Accept': 'application/xml'} doesn't match "
+            "{'Accept': 'application/json'}"
+        ) in msg
+
+    run()
+    assert_reset()
+
+
+def test_request_headers_missing_raises():
+    @responses.activate
+    def run():
+        url = "http://example.com/"
+        responses.add(
+            method=responses.GET,
+            url=url,
+            json={"success": True},
+            match=[matchers.header_matcher({"x-custom-header": "foo"})],
+        )
+
+        with pytest.raises(ConnectionError) as excinfo:
+            requests.get(url, headers={})
+
+        msg = str(excinfo.value)
+        assert (
+            "Headers do not match: {} doesn't match {'x-custom-header': 'foo'}"
         ) in msg
 
     run()
@@ -683,8 +734,8 @@ def test_request_matches_headers_strict_match():
 
         msg = str(excinfo.value)
         assert (
-            "Headers do not match: {Accept: text/plain, Accept-Charset: utf-8} "
-            "doesn't match {Accept: text/plain}"
+            "Headers do not match: {'Accept': 'text/plain', 'Accept-Charset': 'utf-8'} "
+            "doesn't match {'Accept': 'text/plain'}"
         ) in msg
 
     run()
@@ -759,26 +810,165 @@ def test_fragment_identifier_matcher_and_match_querystring():
     assert_reset()
 
 
-def test_matchers_create_key_val_str():
-    """
-    Test that matchers._create_key_val_str does recursive conversion
-    """
-    data = {
-        "my_list": [
-            1,
-            2,
-            "a",
-            {"key1": "val1", "key2": 2, 3: "test"},
-            "!",
-            [["list", "nested"], {"nested": "dict"}],
-        ],
-        1: 4,
-        "test": "val",
-        "high": {"nested": "nested_dict"},
-    }
-    conv_str = matchers._create_key_val_str(data)
-    reference = (
-        "{1: 4, high: {nested: nested_dict}, my_list: [!, 1, 2, [[list, nested], {nested: dict}], "
-        "a, {3: test, key1: val1, key2: 2}], test: val}"
-    )
-    assert conv_str == reference
+class TestHeaderWithRegex:
+    @property
+    def url(self):  # type: ignore[misc]
+        return "http://example.com/"
+
+    def _register(self):
+        responses.add(
+            method=responses.GET,
+            url=self.url,
+            body="success",
+            match=[
+                matchers.header_matcher(
+                    {
+                        "Accept": "text/plain",
+                        "Message-Signature": re.compile(r'signature="\S+",created=\d+'),
+                    },
+                    strict_match=True,
+                )
+            ],
+        )
+
+    def test_request_matches_headers_regex(self):
+        @responses.activate
+        def run():
+            # this one can not use common _register method because different headers
+            responses.add(
+                method=responses.GET,
+                url=self.url,
+                json={"success": True},
+                match=[
+                    matchers.header_matcher(
+                        {
+                            "Message-Signature": re.compile(
+                                r'signature="\S+",created=\d+'
+                            ),
+                            "Authorization": "Bearer API_TOKEN",
+                        },
+                        strict_match=False,
+                    )
+                ],
+            )
+            # the actual request can contain extra headers (requests always adds some itself anyway)
+            resp = requests.get(
+                self.url,
+                headers={
+                    "Message-Signature": 'signature="abc",created=1243',
+                    "Authorization": "Bearer API_TOKEN",
+                },
+            )
+            assert_response(
+                resp, body='{"success": true}', content_type="application/json"
+            )
+
+        run()
+        assert_reset()
+
+    def test_request_matches_headers_regex_strict_match_regex_failed(self):
+        @responses.activate
+        def run():
+            self._register()
+            session = requests.Session()
+            # requests will add some extra headers of its own, so we have to use prepared requests
+            prepped = session.prepare_request(
+                requests.Request(
+                    method="GET",
+                    url=self.url,
+                )
+            )
+            prepped.headers.clear()
+            prepped.headers["Accept"] = "text/plain"
+            prepped.headers["Message-Signature"] = 'signature="123",created=abc'
+            with pytest.raises(ConnectionError) as excinfo:
+                session.send(prepped)
+            msg = str(excinfo.value)
+            assert (
+                "Headers do not match: {'Accept': 'text/plain', 'Message-Signature': "
+                """'signature="123",created=abc'} """
+                "doesn't match {'Accept': 'text/plain', 'Message-Signature': "
+                "re.compile('signature=\"\\\\S+\",created=\\\\d+')}"
+            ) in msg
+
+        run()
+        assert_reset()
+
+    def test_request_matches_headers_regex_strict_match_mismatched_field(self):
+        @responses.activate
+        def run():
+            self._register()
+            # requests will add some extra headers of its own, so we have to use prepared requests
+            session = requests.Session()
+            prepped = session.prepare_request(
+                requests.Request(
+                    method="GET",
+                    url=self.url,
+                )
+            )
+            prepped.headers.clear()
+            prepped.headers["Accept"] = "text/plain"
+            prepped.headers["Accept-Charset"] = "utf-8"
+            # "Accept-Charset" header will fail to match to "Message-Signature"
+            with pytest.raises(ConnectionError) as excinfo:
+                session.send(prepped)
+            msg = str(excinfo.value)
+            assert (
+                "Headers do not match: {'Accept': 'text/plain', 'Accept-Charset': 'utf-8'} "
+                "doesn't match {'Accept': 'text/plain', 'Message-Signature': "
+                "re.compile('signature=\"\\\\S+\",created=\\\\d+')}"
+            ) in msg
+
+        run()
+        assert_reset()
+
+    def test_request_matches_headers_regex_strict_match_mismatched_number(self):
+        @responses.activate
+        def run():
+            self._register()
+            # requests will add some extra headers of its own, so we have to use prepared requests
+            session = requests.Session()
+            # include the "Accept-Charset" header, which will fail to match
+            prepped = session.prepare_request(
+                requests.Request(
+                    method="GET",
+                    url=self.url,
+                )
+            )
+            prepped.headers.clear()
+            prepped.headers["Accept"] = "text/plain"
+            prepped.headers["Accept-Charset"] = "utf-8"
+            prepped.headers["Message-Signature"] = 'signature="abc",created=1243'
+            with pytest.raises(ConnectionError) as excinfo:
+                session.send(prepped)
+            msg = str(excinfo.value)
+            assert (
+                "Headers do not match: {'Accept': 'text/plain', 'Accept-Charset': 'utf-8', "
+                """'Message-Signature': 'signature="abc",created=1243'} """
+                "doesn't match {'Accept': 'text/plain', 'Message-Signature': "
+                "re.compile('signature=\"\\\\S+\",created=\\\\d+')}"
+            ) in msg
+
+        run()
+        assert_reset()
+
+    def test_request_matches_headers_regex_strict_match_positive(self):
+        @responses.activate
+        def run():
+            self._register()
+            # requests will add some extra headers of its own, so we have to use prepared requests
+            session = requests.Session()
+            prepped = session.prepare_request(
+                requests.Request(
+                    method="GET",
+                    url=self.url,
+                )
+            )
+            prepped.headers.clear()
+            prepped.headers["Accept"] = "text/plain"
+            prepped.headers["Message-Signature"] = 'signature="abc",created=1243'
+            resp = session.send(prepped)
+            assert_response(resp, body="success", content_type="text/plain")
+
+        run()
+        assert_reset()
